#pragma once


#include <cstdlib>
#include <cmath>
#include <numbers>
#include <QObject>
#include <QMutex>
#include <QElapsedTimer>
#include <bgtu/computer_graphics_lab_work/renderer_api/point.hpp>
#include <bgtu/computer_graphics_lab_work/renderer_api/sprite.hpp>
#include <bgtu/computer_graphics_lab_work/utilities/matrix.hpp>
#include <bgtu/computer_graphics_lab_work/utilities/default_renderer_linear.hpp>
#include <bgtu/computer_graphics_lab_work/utilities/zoomed_voxel_painter.hpp>


namespace BGTU::ComputerGraphicsLabWork::Impl {
	class _Lab5SpriteData_Provider;


	struct SpriteData {
	public:
		RendererApi::PointI<2> central_pixel;
		double radius_multiplier;
		Utilities::Matrix3d planets_transform[9];
		Utilities::Matrix3d stars_transform;
		double time_ms;
		using Provider = _Lab5SpriteData_Provider;

	};

	class _Lab5SpriteData_Provider : public QObject, public RendererApi::Sprite<SpriteData, Utilities::DefaultVoxelDrawerCache::VoxelPainterImpl>::SpriteDataProvider {
	Q_OBJECT
	private:
		QMutex sync;
		QElapsedTimer time;
		RendererApi::VirtualVoxelPainter::visible_pixel_coordinate_fast_t w, h;

	public:
		explicit _Lab5SpriteData_Provider(double radians_per_second) :
				sync{}, time{},
				w{0}, h{0} {}

		SpriteData get_sprite_data() override {
			this->sync.lock();
			double radius = (((this->w < this->h) ? this->w : this->h) * 7 / 16.0);
			Utilities::Matrix3d center_shift = Utilities::Matrix3d::shift(this->w / 2, this->h / 2);
			Utilities::Matrix3d scale = Utilities::Matrix3d::scale(radius);
			double current_time = this->time.elapsed();
			SpriteData cached{
					.central_pixel = RendererApi::PointI<2>(this->w, this->h) / 2,
					.radius_multiplier = radius,
					.planets_transform = {
							center_shift * Utilities::Matrix3d::scale(radius),
							center_shift * scale * Utilities::Matrix3d::rotate(current_time / 1000.0 * 0.015),
							center_shift * scale * Utilities::Matrix3d::rotate(current_time / 1000.0 * 0.012),
							center_shift * scale * Utilities::Matrix3d::rotate(current_time / 1000.0 * 0.01),
							center_shift * scale * Utilities::Matrix3d::rotate(current_time / 1000.0 * 0.009),
							center_shift * scale * Utilities::Matrix3d::rotate(current_time / 1000.0 * 0.005),
							center_shift * scale * Utilities::Matrix3d::rotate(current_time / 1000.0 * 0.0045),
							center_shift * scale * Utilities::Matrix3d::rotate(current_time / 1000.0 * 0.002),
							center_shift * scale * Utilities::Matrix3d::rotate(current_time / 1000.0 * 0.0017),
					},
					.stars_transform = Utilities::Matrix3d::rotate(current_time / 1000.0 * 0.0002) * scale * center_shift,
					.time_ms = current_time

			};
			this->sync.unlock();
			return cached;
		}

	public slots:

		void set_frame_size(RendererApi::VirtualVoxelPainter::visible_pixel_coordinate_fast_t ww, RendererApi::VirtualVoxelPainter::visible_pixel_coordinate_fast_t hh) {
			this->sync.lock();
			this->w = ww;
			this->h = hh;
			this->sync.unlock();
		}
	};
}