#include <cmath>
#include <array>
#include <iostream>
#include <numbers>
#include <QApplication>
#include <QMainWindow>
#include <QTimer>
#include <QGridLayout>
#include <bgtu/computer_graphics_lab_work/renderer_api/color.hpp>
#include <bgtu/computer_graphics_lab_work/utilities/default_renderer_linear.hpp>
#include <bgtu/computer_graphics_lab_work/utilities/voxel_painter_exporter_to_packed_array.hpp>
#include <bgtu/computer_graphics_lab_work/utilities/matrix.hpp>
#include <bgtu/computer_graphics_lab_work/qt_utilities/separate_threaded_renderer.hpp>
#include <bgtu/computer_graphics_lab_work/qt_utilities/renderer_widget.hpp>
#include <bgtu/computer_graphics_lab_work/qt_utilities/keyboard_catcher_widget.hpp>
#include "sprite_data.hpp"
#include "variant4.hpp"

namespace BGTU::ComputerGraphicsLabWork::Impl {
	int main(int argc, char **argv) {
#if 1
		QApplication qApplication{argc, argv};

		SpriteData::Provider sprites_data_front{0.0, 0.0, 1.01};
		SpriteData::Provider sprites_data_side{std::numbers::pi_v<double> / 2, 0, 1.01};
		sprites_data_side.inc_x_angle();
		SpriteData::Provider sprites_data_up{0, std::numbers::pi_v<double> / 2, 1.01};
		sprites_data_up.inc_y_angle();
		SpriteData::Provider sprites_data_custom{0.1, 0.1, 1.01};

		QMainWindow w{};


		QtUtilities::KeyboardCatcherWidget kbd{&w};

		QObject::connect(&kbd, &QtUtilities::KeyboardCatcherWidget::key_pressed_Left, &sprites_data_custom, &SpriteData::Provider::dec_y_angle);
		QObject::connect(&kbd, &QtUtilities::KeyboardCatcherWidget::key_pressed_Right, &sprites_data_custom, &SpriteData::Provider::inc_y_angle);
		QObject::connect(&kbd, &QtUtilities::KeyboardCatcherWidget::key_pressed_Down, &sprites_data_custom, &SpriteData::Provider::dec_x_angle);
		QObject::connect(&kbd, &QtUtilities::KeyboardCatcherWidget::key_pressed_Up, &sprites_data_custom, &SpriteData::Provider::inc_x_angle);
		QObject::connect(&kbd, &QtUtilities::KeyboardCatcherWidget::key_pressed_A, &sprites_data_custom, &SpriteData::Provider::dec_y_angle);
		QObject::connect(&kbd, &QtUtilities::KeyboardCatcherWidget::key_pressed_D, &sprites_data_custom, &SpriteData::Provider::inc_y_angle);
		QObject::connect(&kbd, &QtUtilities::KeyboardCatcherWidget::key_pressed_S, &sprites_data_custom, &SpriteData::Provider::dec_x_angle);
		QObject::connect(&kbd, &QtUtilities::KeyboardCatcherWidget::key_pressed_W, &sprites_data_custom, &SpriteData::Provider::inc_x_angle);
		QObject::connect(&kbd, &QtUtilities::KeyboardCatcherWidget::key_pressed_PageDown, &sprites_data_front, &SpriteData::Provider::dec_scale);
		QObject::connect(&kbd, &QtUtilities::KeyboardCatcherWidget::key_pressed_PageUp, &sprites_data_front, &SpriteData::Provider::inc_scale);
		QObject::connect(&kbd, &QtUtilities::KeyboardCatcherWidget::key_pressed_PageDown, &sprites_data_side, &SpriteData::Provider::dec_scale);
		QObject::connect(&kbd, &QtUtilities::KeyboardCatcherWidget::key_pressed_PageUp, &sprites_data_side, &SpriteData::Provider::inc_scale);
		QObject::connect(&kbd, &QtUtilities::KeyboardCatcherWidget::key_pressed_PageDown, &sprites_data_up, &SpriteData::Provider::dec_scale);
		QObject::connect(&kbd, &QtUtilities::KeyboardCatcherWidget::key_pressed_PageUp, &sprites_data_up, &SpriteData::Provider::inc_scale);
		QObject::connect(&kbd, &QtUtilities::KeyboardCatcherWidget::key_pressed_PageDown, &sprites_data_custom, &SpriteData::Provider::dec_scale);
		QObject::connect(&kbd, &QtUtilities::KeyboardCatcherWidget::key_pressed_PageUp, &sprites_data_custom, &SpriteData::Provider::inc_scale);



		QtUtilities::SeparateThreadedDefaultRendererLinear<SpriteData> renderer_front{};
		renderer_front.set_sprite_data_provider(&sprites_data_front);
		renderer_front.set_sprites(sprites, sprites_count);
		renderer_front.set_background(BGTU::ComputerGraphicsLabWork::RendererApi::Color{0, 0, 0});
		QtUtilities::SeparateThreadedDefaultRendererLinear<SpriteData> renderer_side{};
		renderer_side.set_sprite_data_provider(&sprites_data_side);
		renderer_side.set_sprites(sprites, sprites_count);
		renderer_side.set_background(BGTU::ComputerGraphicsLabWork::RendererApi::Color{0, 0, 0});
		QtUtilities::SeparateThreadedDefaultRendererLinear<SpriteData> renderer_up{};
		renderer_up.set_sprite_data_provider(&sprites_data_up);
		renderer_up.set_sprites(sprites, sprites_count);
		renderer_up.set_background(BGTU::ComputerGraphicsLabWork::RendererApi::Color{0, 0, 0});
		QtUtilities::SeparateThreadedDefaultRendererLinear<SpriteData> renderer_custom{};
		renderer_custom.set_sprite_data_provider(&sprites_data_custom);
		renderer_custom.set_sprites(sprites, sprites_count);
		renderer_custom.set_background(BGTU::ComputerGraphicsLabWork::RendererApi::Color{0, 0, 0});


		QGridLayout layout{&kbd};
		kbd.setLayout(&layout);

		QtUtilities::RendererWidget<SpriteData> canvas_front{&renderer_front, &kbd};
		layout.addWidget(&canvas_front, 0, 0);
		QObject::connect(&canvas_front, &QtUtilities::_RendererWidget_SignalSlots::resized, &sprites_data_front, &SpriteData::Provider::set_frame_size);
		QtUtilities::RendererWidget<SpriteData> canvas_side{&renderer_side, &kbd};
		layout.addWidget(&canvas_side, 0, 1);
		QObject::connect(&canvas_side, &QtUtilities::_RendererWidget_SignalSlots::resized, &sprites_data_side, &SpriteData::Provider::set_frame_size);
		QtUtilities::RendererWidget<SpriteData> canvas_up{&renderer_up, &kbd};
		layout.addWidget(&canvas_up, 1, 0);
		QObject::connect(&canvas_up, &QtUtilities::_RendererWidget_SignalSlots::resized, &sprites_data_up, &SpriteData::Provider::set_frame_size);
		QtUtilities::RendererWidget<SpriteData> canvas_custom{&renderer_custom, &kbd};
		layout.addWidget(&canvas_custom, 1, 1);
		QObject::connect(&canvas_custom, &QtUtilities::_RendererWidget_SignalSlots::resized, &sprites_data_custom, &SpriteData::Provider::set_frame_size);


		w.setCentralWidget(&kbd);
		w.show();

		QApplication::exec();

		return 0;
#else
		Utilities::Matrix3d x = Utilities::Matrix3d::shift(0, 9);
		Utilities::Matrix3d y = Utilities::Matrix3d::rotate(3.14);

		RendererApi::PointF<2> a{0, 1};
		RendererApi::PointF<2> b = x * a;
		RendererApi::PointF<2> c = x * y * a;
		RendererApi::PointF<2> d = y * x * a;


		return 0;
#endif
	}
}


int main(int argc, char **argv) {
	return BGTU::ComputerGraphicsLabWork::Impl::main(argc, argv);
}