#include <utility>
#include <numbers>
#include <cmath>
#include <bgtu/computer_graphics_lab_work/utilities/shader.hpp>
#include <bgtu/computer_graphics_lab_work/utilities/shapes/triangle.hpp>

#include "triangle.hpp"
#include "sprite_data.hpp"
#include "model.hpp"

namespace BGTU::ComputerGraphicsLabWork::Impl {
	namespace {

		Utilities::hsv_sector_shader hsv_shader{0, 0, 0};
		Utilities::MonoShader transparent_shader{{255, 255, 255, 127}};

		class ChessBoardShader : public Utilities::Shader {
		private:
			double w, h;

		public:
			ChessBoardShader(double w, double h) : w{w}, h{h} {}

			[[nodiscard]] RendererApi::Color::Transparent get_color(double x, double y) const override {
				if (static_cast<long long>(std::floor(x / this->w) + std::floor(y / this->h)) % 2 == 0)
					return {0, 0, 0, 255};
				else
					return {255, 255, 255, 255};
			}
		} chess_board_shader{0.125000000000, 0.125000000000};


		RendererApi::PointF<2> rot_n(double degrees, unsigned len) {
			auto radians = degrees * std::numbers::pi_v<double> / 180;
			return {std::cos(radians) * len, std::sin(radians) * len};
		}

		Utilities::Shapes::triangle_barycentric_shader tri_shader_1{
				{0, 0}, RendererApi::Color{0, 0, 255},
				{1, 0}, RendererApi::Color{0, 255, 255},
				{0, 1}, RendererApi::Color{0, 255, 0}
		};
		Utilities::Shapes::triangle_barycentric_shader tri_shader_2{
				{1, 1}, RendererApi::Color{255, 255, 0},
				{1, 0}, RendererApi::Color{0, 255, 255},
				{0, 1}, RendererApi::Color{0, 255, 0}
		};

		Triangle triangles[] = {
				{{0,  1,  1},  {1,  1,  0},  {1,  0,  1},  rot_n(0, 1),   rot_n(240, 1), rot_n(120, 1), &hsv_shader},
				{{-1, -1, 1},  {-1, 1,  1},  {0,  1,  1},  rot_n(60, 2),  rot_n(0, 2),   rot_n(0, 1),   &hsv_shader},
				{{-1, -1, 1},  {1,  0,  1},  {0,  1,  1},  rot_n(60, 2),  rot_n(120, 1), rot_n(0, 1),   &hsv_shader},
				{{-1, -1, 1},  {1,  0,  1},  {1,  -1, 1},  rot_n(60, 2),  rot_n(120, 1), rot_n(120, 1), &hsv_shader},
				{{1,  -1, -1}, {1,  1,  -1}, {1,  1,  0},  rot_n(180, 2), rot_n(240, 2), rot_n(240, 1), &hsv_shader},
				{{1,  -1, -1}, {1,  0,  1},  {1,  1,  0},  rot_n(180, 2), rot_n(120, 1), rot_n(240, 1), &hsv_shader},
				{{1,  -1, -1}, {1,  0,  1},  {1,  -1, 1},  rot_n(180, 2), rot_n(120, 1), rot_n(120, 2), &hsv_shader},
				{{-1, 1,  -1}, {1,  1,  -1}, {1,  1,  0},  rot_n(300, 2), rot_n(240, 2), rot_n(240, 1), &hsv_shader},
				{{-1, 1,  -1}, {0,  1,  1},  {1,  1,  0},  rot_n(300, 2), rot_n(0, 1),   rot_n(240, 1), &hsv_shader},
				{{-1, 1,  -1}, {0,  1,  1},  {-1, 1,  1},  rot_n(300, 2), rot_n(0, 1),   rot_n(0, 2),   &hsv_shader},
				{{-1, -1, 1},  {-1, -1, -1}, {-1, 1,  1},  {0, 0},        {0, 1},        {1, 0},        &chess_board_shader},
				{{-1, 1,  -1}, {-1, -1, -1}, {-1, 1,  1},  {1, 1},        {0, 1},        {1, 0},        &chess_board_shader},
				{{-1, -1, -1}, {-1, 1,  -1}, {1,  1,  -1}, {0, 0},        {0, 1},        {1, 1},        &transparent_shader},
				{{-1, -1, -1}, {1,  -1, -1}, {1,  1,  -1}, {0, 0},        {1, 0},        {1, 1},        &transparent_shader},
				{{1,  -1, 1},  {-1, -1, 1},  {1,  -1, -1}, {0, 0},        {0, 1},        {1, 0},        &tri_shader_1},
				{{-1, -1, -1}, {-1, -1, 1},  {1,  -1, -1}, {1, 1},        {0, 1},        {1, 0},        &tri_shader_2},
		};

		Triangle *const _sprites[]{
				&triangles[0],
				&triangles[1],
				&triangles[2],
				&triangles[3],
				&triangles[4],
				&triangles[5],
				&triangles[6],
				&triangles[7],
				&triangles[8],
				&triangles[9],
				&triangles[10],
				&triangles[11],
				&triangles[12],
				&triangles[13],
				&triangles[14],
				&triangles[15],
				&triangles[16],
				&triangles[17],
				&triangles[18],
				&triangles[19],
		};
	}


	Triangle *const *const sprites = _sprites;
	std::size_t sprites_count = 16;
}
