#pragma once

#include <algorithm>

#include <bgtu/computer_graphics_lab_work/renderer_api/point.hpp>
#include <bgtu/computer_graphics_lab_work/renderer_api/sprite.hpp>
#include <bgtu/computer_graphics_lab_work/utilities/matrix.hpp>
#include <bgtu/computer_graphics_lab_work/utilities/shader.hpp>
#include <bgtu/computer_graphics_lab_work/utilities/shapes/triangle.hpp>

#include "triangle.hpp"


namespace BGTU::ComputerGraphicsLabWork::Impl {
	class SortSprite : public RendererApi::Sprite<SpriteData> {
	private:
		Triangle const *const *triangles;
		std::size_t triangles_count;
	public:
		SortSprite(
				Triangle const *const *triangles,
				std::size_t triangles_count

		) : triangles{triangles}, triangles_count{triangles_count} {}

		void draw(BGTU::ComputerGraphicsLabWork::RendererApi::VirtualVoxelPainter *frame, SpriteData const *data) const override {
			auto indices = new std::size_t[this->triangles_count];
			for (std::size_t i = this->triangles_count; i-- > 0;) {
				indices[i] = i;
			}

			std::sort(indices, indices + this->triangles_count, [this](std::size_t l, std::size_t r) -> bool {
				return (this->triangles[l]->triangle_p0.z + this->triangles[l]->triangle_p1.z + this->triangles[l]->triangle_p2.z) / 3 > (this->triangles[r]->triangle_p0.z + this->triangles[r]->triangle_p1.z + this->triangles[r]->triangle_p2.z) / 3;
			});

			for (std::size_t i = 0; i < this->triangles_count; i++) {
				this->triangles[indices[i]]->draw(frame, data);
			}
			delete[] indices;
		}
	};
}