#pragma once

#include <bgtu/computer_graphics_lab_work/renderer_api/point.hpp>
#include <bgtu/computer_graphics_lab_work/renderer_api/sprite.hpp>
#include <bgtu/computer_graphics_lab_work/utilities/matrix.hpp>
#include <bgtu/computer_graphics_lab_work/utilities/shader.hpp>
#include <bgtu/computer_graphics_lab_work/utilities/shapes/triangle.hpp>

#include "sprite_data.hpp"


namespace BGTU::ComputerGraphicsLabWork::Impl {
	class Triangle : public RendererApi::Sprite<SpriteData> {
	public:
		const RendererApi::PointF<3> triangle_p0;
		const RendererApi::PointF<3> triangle_p1;
		const RendererApi::PointF<3> triangle_p2;

	private:
		const RendererApi::PointF<2> shader_p0;
		const RendererApi::PointF<2> shader_p1;
		const RendererApi::PointF<2> shader_p2;
		Utilities::Shader const *const shader;

	public:
		Triangle(
				RendererApi::PointF<3> triangle_p0,
				RendererApi::PointF<3> triangle_p1,
				RendererApi::PointF<3> triangle_p2,
				RendererApi::PointF<2> shader_p0,
				RendererApi::PointF<2> shader_p1,
				RendererApi::PointF<2> shader_p2,
				Utilities::Shader const *shader
		) : triangle_p0{triangle_p0}, triangle_p1{triangle_p1}, triangle_p2{triangle_p2},
			shader_p0{shader_p0}, shader_p1{shader_p1}, shader_p2{shader_p2}, shader{shader} {}

		void draw(BGTU::ComputerGraphicsLabWork::RendererApi::VirtualVoxelPainter *frame, SpriteData const *data) const override {
			auto ta4 = data->transform * Utilities::Vec4{this->triangle_p0.x, this->triangle_p0.y, -this->triangle_p0.z, 1.0};
			auto tb4 = data->transform * Utilities::Vec4{this->triangle_p1.x, this->triangle_p1.y, -this->triangle_p1.z, 1.0};
			auto tc4 = data->transform * Utilities::Vec4{this->triangle_p2.x, this->triangle_p2.y, -this->triangle_p2.z, 1.0};

			RendererApi::PointF<3> ta{ta4[0] / ta4[3], ta4[1] / ta4[3], ta4[2] / ta4[3]};
			RendererApi::PointF<3> tb{tb4[0] / tb4[3], tb4[1] / tb4[3], tb4[2] / tb4[3]};
			RendererApi::PointF<3> tc{tc4[0] / tc4[3], tc4[1] / tc4[3], tc4[2] / tc4[3]};

#if 1
			Utilities::Shapes::triangle_perspective_interpolator<RendererApi::PointF<3>> interpolator{
					{ta.x,              ta.y, ta4[3]},
					{this->shader_p0.x, this->shader_p0.y, ta.z},
					{tb.x,              tb.y, tb4[3]},
					{this->shader_p1.x, this->shader_p1.y, tb.z},
					{tc.x,              tc.y, tc4[3]},
					{this->shader_p2.x, this->shader_p2.y, tc.z},
			};
#else
			Utilities::Shapes::triangle_barycentric_interpolator<RendererApi::PointF<3>> interpolator{
					{ta.x,              ta.y},
					{this->shader_p0.x, this->shader_p0.y, ta.z},
					{tb.x,              tb.y},
					{this->shader_p1.x, this->shader_p1.y, tb.z},
					{tc.x,              tc.y},
					{this->shader_p2.x, this->shader_p2.y, tc.z},
			};
#endif

			Utilities::Shapes::iterate_triangle_fill(
					{static_cast<RendererApi::PointI<2>::component_t>(ta.x), static_cast<RendererApi::PointI<2>::component_t>(ta.y)},
					{static_cast<RendererApi::PointI<2>::component_t>(tb.x), static_cast<RendererApi::PointI<2>::component_t>(tb.y)},
					{static_cast<RendererApi::PointI<2>::component_t>(tc.x), static_cast<RendererApi::PointI<2>::component_t>(tc.y)},
					[&, this](bool is_edge, RendererApi::PointI<2>::component_t x, RendererApi::PointI<2>::component_t y) {
						RendererApi::PointF<3> _tmp = interpolator.interpolate_point(x, y);
						frame->add_voxel(x, y, _tmp.z, this->shader->get_color(_tmp.x, _tmp.y));
					}
			);

		}
	};
}