#pragma once

#include <cassert>
#include <cstring>
#include <Qt>
#include <QObject>
#include <QImage>
#include <QColor>
#include <bgtu/computer_graphics_lab_work/renderer_api/color.hpp>
#include <bgtu/computer_graphics_lab_work/renderer_api/point.hpp>
#include <bgtu/computer_graphics_lab_work/renderer_api/voxel_painter.hpp>
#include <bgtu/computer_graphics_lab_work/renderer_api/sprite.hpp>

namespace BGTU::ComputerGraphicsLabWork::QtUtilities {
	class OwnedQImage : public QImage, public RendererApi::VirtualVoxelPainter::Exporter {
	private:
		RendererApi::SpriteMetadata **owners;
		unsigned char *pixels_buffer;
		bool owns_buffers;

	public:
		OwnedQImage(unsigned width, unsigned height) : QImage(width, height, QImage::Format_ARGB32), pixels_buffer{nullptr}, owns_buffers{true} {
			if (width == 0 || height == 0) {
				this->owners = nullptr;
			} else {
				this->owners = new RendererApi::SpriteMetadata *[width * height];
				for (RendererApi::SpriteMetadata **p = this->owners - 1; p-- >= this->owners;) {
					*p = nullptr;
				}
			}
		}

		OwnedQImage() : QImage(), owners{nullptr}, pixels_buffer{nullptr} {}

		OwnedQImage(std::uint_least8_t *data, unsigned width, unsigned height, QImage::Format f, RendererApi::SpriteMetadata **owners, QImageCleanupFunction destructor, void *destructor_closure) :
				QImage(data, width, height, f, destructor, destructor_closure), pixels_buffer{data}, owners{owners}, owns_buffers{false} {}

		OwnedQImage(std::uint_least8_t *data, unsigned width, unsigned height, QImage::Format f, RendererApi::SpriteMetadata **owners) :
				QImage(data, width, height, f), pixels_buffer{data}, owners{owners}, owns_buffers{false} {}

		enum _DelegateBuffers {
			DelegateBuffers
		};

		OwnedQImage(std::uint_least8_t *data, unsigned width, unsigned height, QImage::Format f, RendererApi::SpriteMetadata **owners, _DelegateBuffers) :
				QImage(data, width, height, f), pixels_buffer{data}, owners{owners}, owns_buffers{true} {}

#if 0
		OwnedQImage &operator=(OwnedQImage const &copy) {
			if (this == &copy)
				return *this;
			if (this->owners != nullptr)
				delete[] this->owners;
			if (copy.owners == nullptr) {
				this->owners = nullptr;
			} else {
				this->owners = new RendererApi::SpriteMetadata const *[copy.width() * copy.height()];
				std::memcpy(this->owners, copy.owners, copy.width() * copy.height() * sizeof(RendererApi::SpriteMetadata const *));
			}
			return *this;
		}

		OwnedQImage(OwnedQImage const &copy) : QImage{copy} {
			*this = copy;
		}

		OwnedQImage(OwnedQImage &&move) noexcept: QImage{std::move(move)} {
			if (this->owners != nullptr)
				delete[] this->owners;
			this->owners = move.owners;
			move.owners = nullptr;
		}
#endif

		void setPixelOwner(unsigned x, unsigned y, RendererApi::SpriteMetadata *o) {
			assert(x < this->width());
			assert(y < this->height());

			this->owners[y * this->width() + x] = o;
		}

		[[nodiscard]] RendererApi::SpriteMetadata *getPixelOwner(unsigned x, unsigned y) const {
			assert(x < this->width());
			assert(y < this->height());

			return this->owners[y * this->width() + x];
		}

		void set_pixel(RendererApi::PointI<2>::component_t x, RendererApi::PointI<2>::component_t y, RendererApi::Color c, RendererApi::SpriteMetadata *owner) final {
			this->setPixelColor(x, y, QColor{c.red, c.green, c.blue});
			this->setPixelOwner(x, y, owner);
		}

		~OwnedQImage() override {
			if (this->owns_buffers) {
				if (this->owners != nullptr) {
					delete[] this->owners;
					this->owners = nullptr;
				}
				if (this->pixels_buffer != nullptr) {
					delete[] this->pixels_buffer;
					this->pixels_buffer = nullptr;
				}
			}
		}
	};
}