#pragma once

#include <utility>
#include <Qt>
#include <QObject>
#include <QWidget>
#include <QMutex>
#include <QPainter>
#include <QBrush>
#include <QApplication>
#include <QPalette>
#include <QStyle>
#include <QPaintEvent>
#include <QResizeEvent>
#include <QMouseEvent>
#include <bgtu/computer_graphics_lab_work/utilities/default_renderer_linear.hpp>
#include "owned_qimage.hpp"
#include "separate_threaded_renderer.hpp"
#include "_renderer_widget.hpp"


namespace BGTU::ComputerGraphicsLabWork::QtUtilities {
	template<class sprite_data_t>
	class RendererWidget : public _RendererWidget_SignalSlots {
	private:
		QMutex sync;
		OwnedQImage *next_image;
		OwnedQImage *current_image;
		_SeparateThreadedRenderer_Signals *renderer;

	private slots:

		void receive_image(BGTU::ComputerGraphicsLabWork::QtUtilities::OwnedQImage *img) final {
			this->sync.lock();
			if (this->next_image != nullptr && this->next_image != img)
				delete this->next_image;
			this->next_image = img;
			this->sync.unlock();
			this->update();
		}


	public:
		explicit RendererWidget(_SeparateThreadedRenderer_Signals *renderer, QWidget *owner = nullptr) : _RendererWidget_SignalSlots{owner}, sync{}, next_image{nullptr}, current_image{nullptr}, renderer{renderer} {
			QObject::connect(this->renderer, &_SeparateThreadedRenderer_Signals::frame_rendered, this, &_RendererWidget_SignalSlots::receive_image);
		}

	protected:
		void paintEvent(QPaintEvent *event) final {
			this->sync.lock();
			if (this->next_image != nullptr) {
				if (this->current_image != nullptr && this->next_image != this->current_image)
					delete this->current_image;
				this->current_image = this->next_image;
				this->next_image = nullptr;
			}
			this->sync.unlock();

			QPainter painter;
			painter.begin(this);
			painter.setPen(Qt::NoPen);
			painter.setBrush(QApplication::style()->standardPalette().brush(QPalette::Background));
			painter.drawRect(this->rect().x(), this->rect().y(), this->rect().width() - 1, this->rect().height() - 1);
			if (this->current_image != nullptr)
				painter.drawImage(this->current_image->rect(), *this->current_image);
			painter.end();
		}

		void resizeEvent(QResizeEvent *event) override {
			this->renderer->set_frame_size(event->size().width(), event->size().height());
			emit this->resized(event->size().width(), event->size().height());
		}

		void mousePressEvent(QMouseEvent *event) override {
			this->sync.lock();
			auto sprite = this->current_image->getPixelOwner(event->x(), event->y());
			this->sync.unlock();
			if (sprite != nullptr)
				sprite->clicked();
		}
	};
}