#pragma once

#include <cstdint>
#include <type_traits>

namespace BGTU::ComputerGraphicsLabWork::RendererApi {
	template<unsigned DIMENSIONS, class _component_t>
	class Point {
		static_assert(DIMENSIONS == 2 || DIMENSIONS == 3);
	};

	template<class _component_t>
	class Point<2, _component_t> {
	public:
		using component_t = _component_t;

		component_t x;
		component_t y;

		constexpr Point(component_t x, component_t y) : x{x}, y{y} {}

		[[nodiscard]]    constexpr Point<2, _component_t> operator+(Point<2, _component_t> const &other) const noexcept {
			return Point<2, _component_t>{this->x + other.x, this->y + other.y};
		}

		[[nodiscard]] constexpr Point<2, _component_t> operator-(Point<2, _component_t> const &other) const noexcept {
			return Point<2, _component_t>{this->x - other.x, this->y - other.y};
		}

		[[nodiscard]] constexpr Point<2, _component_t> operator*(_component_t const &other) const noexcept {
			return Point<2, _component_t>{this->x * other, this->y * other};
		}

		[[nodiscard]] constexpr Point<2, _component_t> operator/(_component_t const &other) const noexcept {
			return Point<2, _component_t>{this->x / other, this->y / other};
		}

		template<class _self_component_t = _component_t, class sfinae = std::enable_if<std::is_same_v<_self_component_t, long long>, void>::type>
		[[nodiscard]] constexpr explicit operator Point<2, double>() const noexcept {
			return Point<2, double>{(double) this->x, (double) this->y};
		}

		template<class _self_component_t = _component_t, class sfinae = std::enable_if<std::is_same_v<_self_component_t, double>, void>::type>
		[[nodiscard]] constexpr explicit operator Point<2, long long>() const noexcept {
			return Point<2, long long>{(long long) this->x, (long long) this->y};
		}

		[[nodiscard]] constexpr bool operator==(Point<2, _component_t> const &other) const noexcept {
			return this->x == other.x && this->y == other.y;
		}

		[[nodiscard]] constexpr bool operator!=(Point<2, _component_t> const &other) const noexcept {
			return this->x != other.x || this->y != other.y;
		}
	};

	template<class _component_t>
	class Point<3, _component_t> {
	public:
		using component_t = _component_t;

		component_t x;
		component_t y;
		component_t z;

		constexpr Point(component_t x, component_t y, component_t z) : x{x}, y{y}, z{z} {}


		template<class new_component_t>
		explicit operator Point<3, new_component_t>() const {
			return Point<3, new_component_t>{new_component_t{this->x}, new_component_t{this->y}, new_component_t{this->z}};
		}


		constexpr Point<3, _component_t> operator+(Point<2, _component_t> const &other) const {
			return Point<3, _component_t>{this->x + other.x, this->y + other.y, this->z + other.z};
		}

		constexpr Point<3, _component_t> operator-(Point<2, _component_t> const &other) const {
			return Point<3, _component_t>{this->x - other.x, this->y - other.y, this->z - other.z};
		}

		constexpr Point<3, _component_t> operator*(_component_t const &other) const {
			return Point<3, _component_t>{this->x * other, this->y * other, this->z * other};
		}

		constexpr Point<3, _component_t> operator/(_component_t const &other) const {
			return Point<3, _component_t>{this->x / other, this->y / other, this->z / other};
		}

		template<class _self_component_t = _component_t, class sfinae = std::enable_if<std::is_same_v<_self_component_t, long long>, void>::type>
		[[nodiscard]] constexpr explicit operator Point<3, double>() const noexcept {
			return Point<3, double>{(double) this->x, (double) this->y, (double) this->z};
		}

		template<class _self_component_t = _component_t, class sfinae = std::enable_if<std::is_same_v<_self_component_t, double>, void>::type>
		[[nodiscard]] constexpr explicit operator Point<3, long long>() const noexcept {
			return Point<3, long long>{(long long) this->x, (long long) this->y, (long long) this->z};
		}
	};

	template<unsigned DIMENSIONS>
	using PointI = Point<DIMENSIONS, long long>;

	template<unsigned DIMENSIONS>
	using PointF = Point<DIMENSIONS, double>;
}