#pragma once

#include <cstdint>
#include <limits>

#include "color.hpp"
#include "point.hpp"

namespace BGTU::ComputerGraphicsLabWork::RendererApi {
	class SpriteMetadata;

	class VirtualVoxelPainter {
	public:
		using visible_pixel_coordinate_fast_t = std::uint_fast16_t;
		using visible_pixel_coordinate_compact_t = std::uint_least16_t;

		static constexpr visible_pixel_coordinate_fast_t visible_pixel_min_x = 0;
		static constexpr visible_pixel_coordinate_fast_t visible_pixel_min_y = 0;
		static constexpr visible_pixel_coordinate_fast_t visible_pixel_max_x = std::numeric_limits<visible_pixel_coordinate_compact_t>::max();
		static constexpr visible_pixel_coordinate_fast_t visible_pixel_max_y = std::numeric_limits<visible_pixel_coordinate_compact_t>::max();


		[[nodiscard]] virtual visible_pixel_coordinate_fast_t width() const = 0;

		[[nodiscard]] virtual visible_pixel_coordinate_fast_t height() const = 0;

		virtual void add_voxel(PointI<2>::component_t x, PointI<2>::component_t y, double z, Color::Transparent c) = 0;

		virtual void add_voxel(PointI<2>::component_t x, PointI<2>::component_t y, double z, Color::Transparent c, SpriteMetadata *owner) = 0;

		inline void add_voxel(PointI<2> p, double z, Color::Transparent c) {
			return this->add_voxel(p.x, p.y, z, c);
		}

		inline void add_voxel(PointI<2> p, double z, Color::Transparent c, SpriteMetadata *owner) {
			return this->add_voxel(p.x, p.y, z, c, owner);
		}

		class Exporter {
		public:
			virtual void set_pixel(PointI<2>::component_t x, PointI<2>::component_t y, Color c, SpriteMetadata *owner) = 0;
		};
	};
}