#pragma once

#include <cstdlib>
#include <bgtu/computer_graphics_lab_work/renderer_api/point.hpp>
#include <bgtu/computer_graphics_lab_work/renderer_api/color.hpp>
#include <bgtu/computer_graphics_lab_work/renderer_api/voxel_painter.hpp>

namespace BGTU::ComputerGraphicsLabWork::Utilities::Shapes {
	template<class edger_receiver_t, class fill_receiver_t>
	void iterate_circle_fill_h_lines(
			RendererApi::PointI<2>::component_t cx, RendererApi::PointI<2>::component_t cy,
			RendererApi::PointI<2>::component_t r,
			edger_receiver_t edge_point_receiver,
			fill_receiver_t fill_line_receiver
	) {
		if (r == 0) {
			edge_point_receiver(cx, cy);
			return;
		}
		RendererApi::PointI<2>::component_t x = 0, y = r;
		RendererApi::PointI<2>::component_t last_y = r + 1;
		RendererApi::PointI<2>::component_t _r2 = r * r;
		while (true) {
			RendererApi::PointI<2>::component_t _x2 = x * x;
			RendererApi::PointI<2>::component_t _y_1 = y - 1;
			RendererApi::PointI<2>::component_t D1 = _x2 + y * y - _r2;
			RendererApi::PointI<2>::component_t D2 = _x2 + _y_1 * _y_1 - _r2;

			if (D1 > -D2)
				y--;

			if (x == y) {
				edge_point_receiver(cx - x, cy + y);
				fill_line_receiver(cy + y, cx - x + 1, cx + x);
				edge_point_receiver(cx + x, cy + y);

				edge_point_receiver(cx - x, cy - y);
				fill_line_receiver(cy - y, cx - x + 1, cx + x);
				edge_point_receiver(cx + x, cy - y);
				break;
			} else if (x > y) {
				break;
			}

			if (y < last_y) {
				edge_point_receiver(cx - x, cy + y);
				fill_line_receiver(cy + y, cx - x + 1, cx + x);
				edge_point_receiver(cx + x, cy + y);

				edge_point_receiver(cx - y, cy + x);
				fill_line_receiver(cy + x, cx - y + 1, cx + y);
				edge_point_receiver(cx + y, cy + x);

				if (x > 0) {
					edge_point_receiver(cx - y, cy - x);
					fill_line_receiver(cy - x, cx - y + 1, cx + y);
					edge_point_receiver(cx + y, cy - x);
				}

				edge_point_receiver(cx - x, cy - y);
				fill_line_receiver(cy - y, cx - x + 1, cx + x);
				edge_point_receiver(cx + x, cy - y);

				last_y = y;
			} else {
				edge_point_receiver(cx - x, cy + y);
				edge_point_receiver(cx + x, cy + y);

				edge_point_receiver(cx - y, cy + x);
				fill_line_receiver(cy + x, cx - y + 1, cx + y);
				edge_point_receiver(cx + y, cy + x);

				if (x > 0) {
					edge_point_receiver(cx - y, cy - x);
					fill_line_receiver(cy - x, cx - y + 1, cx + y);
					edge_point_receiver(cx + y, cy - x);
				}

				edge_point_receiver(cx - x, cy - y);
				edge_point_receiver(cx + x, cy - y);
			}
			x++;
		}
	}

	template<class edger_receiver_t, class fill_receiver_t>
	void iterate_circle_fill_h_lines(
			RendererApi::PointI<2> c,
			RendererApi::PointI<2>::component_t r,
			edger_receiver_t edge_point_receiver,
			fill_receiver_t fill_line_receiver
	) {
		iterate_circle_fill_h_lines(c.x, c.y, r, edge_point_receiver, fill_line_receiver);
	}

	template<class receiver_t>
	void iterate_circle_fill(
			RendererApi::PointI<2>::component_t cx, RendererApi::PointI<2>::component_t cy,
			RendererApi::PointI<2>::component_t r,
			receiver_t receiver
	) {
		iterate_circle_fill_h_lines(
				cx, cy, r,
				[&](RendererApi::PointI<2>::component_t x, RendererApi::PointI<2>::component_t y) { receiver(true, x, y); },
				[&](RendererApi::PointI<2>::component_t y, RendererApi::PointI<2>::component_t x1, RendererApi::PointI<2>::component_t x2) {
					for (RendererApi::PointI<2>::component_t x = x1; x < x2; x++) receiver(false, x, y);
				}
		);
	}

	template<class receiver_t>
	void iterate_circle_fill(
			RendererApi::PointI<2> c,
			RendererApi::PointI<2>::component_t r,
			receiver_t receiver
	) {
		iterate_circle_fill<receiver_t>(c.x, c.y, r, receiver);
	}

	template<class receiver_t>
	void iterate_circle_edge(
			RendererApi::PointI<2>::component_t cx, RendererApi::PointI<2>::component_t cy,
			RendererApi::PointI<2>::component_t r,
			receiver_t receiver
	) {
		iterate_circle_fill_h_lines(
				cx, cy, r,
				receiver,
				[](RendererApi::PointI<2>::component_t y, RendererApi::PointI<2>::component_t x1, RendererApi::PointI<2>::component_t x2) {}
		);
	}


	template<class receiver_t>
	void iterate_circle_edge(
			RendererApi::PointI<2> center,
			RendererApi::PointI<2>::component_t r,
			receiver_t receiver
	) {
		return iterate_circle_edge<receiver_t>(center.x, center.y, r, receiver);
	}


	template<class voxel_painter_t = RendererApi::VirtualVoxelPainter>
	void draw_circle_edge(
			voxel_painter_t *painter,
			RendererApi::PointI<2>::component_t cx, RendererApi::PointI<2>::component_t cy,
			RendererApi::PointI<2>::component_t r,
			double z,
			RendererApi::Color::Transparent outline
	) {
		iterate_circle_edge(
				cx, cy, r,
				[=](RendererApi::PointI<2>::component_t x, RendererApi::PointI<2>::component_t y) { painter->add_voxel(x, y, z, outline); }
		);
	}

	template<class voxel_painter_t = RendererApi::VirtualVoxelPainter>
	void draw_circle_edge(
			voxel_painter_t *painter,
			RendererApi::PointI<2> center,
			RendererApi::PointI<2>::component_t r,
			double z,
			RendererApi::Color::Transparent outline
	) {
		iterate_circle_edge(
				center, r,
				[=](RendererApi::PointI<2>::component_t x, RendererApi::PointI<2>::component_t y) { painter->add_voxel(x, y, z, outline); }
		);
	}

	template<class voxel_painter_t = RendererApi::VirtualVoxelPainter>
	void draw_circle_edge(
			voxel_painter_t *painter,
			RendererApi::PointI<2>::component_t cx, RendererApi::PointI<2>::component_t cy,
			RendererApi::PointI<2>::component_t r,
			double z,
			RendererApi::Color::Transparent outline,
			RendererApi::SpriteMetadata *owner
	) {
		iterate_circle_edge(
				cx, cy, r,
				[=](RendererApi::PointI<2>::component_t x, RendererApi::PointI<2>::component_t y) { painter->add_voxel(x, y, z, outline, owner); }
		);
	}

	template<class voxel_painter_t = RendererApi::VirtualVoxelPainter>
	void draw_circle_edge(
			voxel_painter_t *painter,
			RendererApi::PointI<2> center,
			RendererApi::PointI<2>::component_t r,
			double z,
			RendererApi::Color::Transparent outline,
			RendererApi::SpriteMetadata *owner
	) {
		iterate_circle_edge(
				center, r,
				[=](RendererApi::PointI<2>::component_t x, RendererApi::PointI<2>::component_t y) { painter->add_voxel(x, y, z, outline, owner); }
		);
	}
}