#pragma once

#include <cstdlib>
#include <type_traits>
#include <utility>
#include <limits>
#include <bgtu/computer_graphics_lab_work/renderer_api/point.hpp>
#include "line.hpp"
#include "../shader.hpp"

namespace BGTU::ComputerGraphicsLabWork::Utilities::Shapes {
	template<class edge_point_receiver_t, class fill_line_receiver_t>
	void iterate_triangle_fill_h_lines(
			RendererApi::PointI<2>::component_t x0, RendererApi::PointI<2>::component_t y0,
			RendererApi::PointI<2>::component_t x1, RendererApi::PointI<2>::component_t y1,
			RendererApi::PointI<2>::component_t x2, RendererApi::PointI<2>::component_t y2,
			edge_point_receiver_t edge_point_receiver,
			fill_line_receiver_t fill_line_receiver
	) {
		auto _swap = [](RendererApi::PointI<2>::component_t *sx0, RendererApi::PointI<2>::component_t *sy0, RendererApi::PointI<2>::component_t *sx1, RendererApi::PointI<2>::component_t *sy1) {
			RendererApi::PointI<2>::component_t tmp;
			(tmp = *sx0), (*sx0 = *sx1), (*sx1 = tmp);
			(tmp = *sy0), (*sy0 = *sy1), (*sy1 = tmp);
		};

		if (y0 < y1)
			_swap(&x0, &y0, &x1, &y1);
		if (y1 < y2)
			_swap(&x1, &y1, &x2, &y2);
		if (y0 < y1)
			_swap(&x0, &y0, &x1, &y1);
		if (y0 == y1 && x0 < x1 && y0 != y2)
			_swap(&x0, &y0, &x1, &y1);
		if (y1 == y2 && x2 < x1)
			_swap(&x1, &y1, &x2, &y2);


		auto _impl = [&]<bool is_r2l>(
				brezenham_line_iterable::iterator upper_cathetus,
				brezenham_line_iterable::iterator lower_cathetus,
				brezenham_line_iterable::iterator hypotenuse
		) {
			RendererApi::PointI<2>::component_t x, ex, y;

			x = upper_cathetus->x;
			y = upper_cathetus->y;
			edge_point_receiver(x, y);

			UPPER_CATHETUS:
			upper_cathetus++;
			if (upper_cathetus == nullptr) {
				goto LOWER_CATHETUS;
			}
			if (upper_cathetus->y < y) {
				while (true) {
					if constexpr (is_r2l) {
						if (hypotenuse->x < x) {
							edge_point_receiver(hypotenuse->x, hypotenuse->y);
							ex = hypotenuse->x;
							goto HYPOTENUSE_UPPER;
						}
					} else {
						if (hypotenuse->x > x) {
							edge_point_receiver(hypotenuse->x, hypotenuse->y);
							ex = hypotenuse->x;
							goto HYPOTENUSE_UPPER;
						}
					}

					hypotenuse++;
					if (hypotenuse->y < y) {
						edge_point_receiver(upper_cathetus->x, upper_cathetus->y);
						x = upper_cathetus->x;
						y = upper_cathetus->y;
						goto UPPER_CATHETUS;
					}
				}
			}

			edge_point_receiver(upper_cathetus->x, upper_cathetus->y);
			if constexpr (is_r2l) {
				x = ((x < upper_cathetus->x) ? x : (upper_cathetus->x));
			} else {
				x = ((x > upper_cathetus->x) ? x : (upper_cathetus->x));
			}

			goto UPPER_CATHETUS;

			HYPOTENUSE_UPPER:
			hypotenuse++;
			if (hypotenuse->y < y) {
				if constexpr (is_r2l) {
					fill_line_receiver(y, ex + 1, x);
				} else {
					fill_line_receiver(y, x + 1, ex);
				}
				y--;
				x = upper_cathetus->x;
				edge_point_receiver(upper_cathetus->x, y);
				goto UPPER_CATHETUS;
			}

			if constexpr (is_r2l) {
				ex = ((ex > hypotenuse->x) ? ex : (hypotenuse->x));
				if (hypotenuse->x < x)
					edge_point_receiver(hypotenuse->x, hypotenuse->y);
			} else {
				ex = ((ex < hypotenuse->x) ? ex : (hypotenuse->x));
				if (hypotenuse->x > x)
					edge_point_receiver(hypotenuse->x, hypotenuse->y);
			}
			goto HYPOTENUSE_UPPER;

			LOWER_CATHETUS:
			lower_cathetus++;
			if (lower_cathetus == nullptr) {
				ex = x;
				goto HYPOTENUSE_LOWER;
			}
			if (lower_cathetus->y < y) {
				while (true) {
					if constexpr (is_r2l) {
						if (hypotenuse->x < x) {
							edge_point_receiver(hypotenuse->x, hypotenuse->y);
							ex = hypotenuse->x;
							goto HYPOTENUSE_LOWER;
						}
					} else {
						if (hypotenuse->x > x) {
							edge_point_receiver(hypotenuse->x, hypotenuse->y);
							ex = hypotenuse->x;
							goto HYPOTENUSE_LOWER;
						}
					}

					hypotenuse++;
					if (hypotenuse->y < y) {
						edge_point_receiver(lower_cathetus->x, lower_cathetus->y);
						x = lower_cathetus->x;
						y = lower_cathetus->y;
						goto LOWER_CATHETUS;
					}
				}
			}

			edge_point_receiver(lower_cathetus->x, lower_cathetus->y);
			if constexpr (is_r2l) {
				x = ((x < lower_cathetus->x) ? x : (lower_cathetus->x));
			} else {
				x = ((x > lower_cathetus->x) ? x : (lower_cathetus->x));
			}
			goto LOWER_CATHETUS;

			HYPOTENUSE_LOWER:
			hypotenuse++;
			if (hypotenuse == nullptr) {
				return;
			}
			if (hypotenuse->y < y) {
				if constexpr (is_r2l) {
					fill_line_receiver(y, ex + 1, x);
				} else {
					fill_line_receiver(y, x + 1, ex);
				}
				y--;
				x = lower_cathetus->x;
				edge_point_receiver(lower_cathetus->x, y);
				goto LOWER_CATHETUS;
			}
			if constexpr (is_r2l) {
				ex = ((ex > hypotenuse->x) ? ex : (hypotenuse->x));
				if (hypotenuse->x < x)
					edge_point_receiver(hypotenuse->x, hypotenuse->y);
			} else {
				ex = ((ex < hypotenuse->x) ? ex : (hypotenuse->x));
				if (hypotenuse->x > x)
					edge_point_receiver(hypotenuse->x, hypotenuse->y);
			}
			goto HYPOTENUSE_LOWER;
		};

		auto dir = (x0 - x1) * (y2 - y1) - (x2 - x1) * (y0 - y1);
		if (dir > 0) {
			_impl.template operator()<true>(
					brezenham_line_iterable{x0, y0, x1, y1}.begin(),
					brezenham_line_iterable{x1, y1, x2, y2}.begin(),
					brezenham_line_iterable{x0, y0, x2, y2}.begin()
			);
		} else if (dir < 0) {
			_impl.template operator()<false>(
					brezenham_line_iterable{x0, y0, x1, y1}.begin(),
					brezenham_line_iterable{x1, y1, x2, y2}.begin(),
					brezenham_line_iterable{x0, y0, x2, y2}.begin()
			);
		} else {
			iterate_line(x0, y0, x2, y2, edge_point_receiver);
		}
	}

	template<class edge_point_receiver_t, class fill_line_receiver_t>
	void iterate_triangle_fill_h_lines(
			RendererApi::PointI<2> p0,
			RendererApi::PointI<2> p1,
			RendererApi::PointI<2> p2,
			edge_point_receiver_t edge_point_receiver,
			fill_line_receiver_t fill_line_receiver
	) {
		iterate_triangle_fill_h_lines(
				p0.x, p0.y,
				p1.x, p1.y,
				p2.x, p2.y,
				edge_point_receiver,
				fill_line_receiver
		);
	}

	template<class receiver_t>
	void iterate_triangle_fill(
			RendererApi::PointI<2>::component_t x0, RendererApi::PointI<2>::component_t y0,
			RendererApi::PointI<2>::component_t x1, RendererApi::PointI<2>::component_t y1,
			RendererApi::PointI<2>::component_t x2, RendererApi::PointI<2>::component_t y2,
			receiver_t receiver
	) {
		iterate_triangle_fill_h_lines(
				x0, y0,
				x1, y1,
				x2, y2,
				[&](RendererApi::PointI<2>::component_t x, RendererApi::PointI<2>::component_t y) { receiver(true, x, y); },
				[&](RendererApi::PointI<2>::component_t y, RendererApi::PointI<2>::component_t x, RendererApi::PointI<2>::component_t ex) {
					for (; x < ex; x++)
						receiver(false, x, y);
				}
		);
	}

	template<class receiver_t>
	void iterate_triangle_fill(
			RendererApi::PointI<2> p0,
			RendererApi::PointI<2> p1,
			RendererApi::PointI<2> p2,
			receiver_t receiver
	) {
		iterate_triangle_fill(
				p0.x, p0.y,
				p1.x, p1.y,
				p2.x, p2.y,
				receiver
		);
	}

	template<class variable_t, auto transform>
	struct _triangle_barycentric_interpolator {
	private:
		long double full_square;
		RendererApi::PointF<2> p0;
		RendererApi::PointF<2> p1;
		RendererApi::PointF<2> p2;
		variable_t v0;
		variable_t v1;
		variable_t v2;

		static long double calculate_square_2(
				RendererApi::PointF<2> p0,
				RendererApi::PointF<2> p1,
				RendererApi::PointF<2> p2
		) noexcept {
			return (p1.x - p0.x) * (p2.y - p0.y) - (p2.x - p0.x) * (p1.y - p0.y);
		}

	public:
		_triangle_barycentric_interpolator(
				RendererApi::PointF<2> p0, variable_t v0,
				RendererApi::PointF<2> p1, variable_t v1,
				RendererApi::PointF<2> p2, variable_t v2
		) : p0{p0}, p1{p1}, p2{p2}, v0{v0}, v1{v1}, v2{v2}, full_square{calculate_square_2(p0, p1, p2)} {
		}

		[[nodiscard]] variable_t interpolate_point(RendererApi::PointF<2> p) const {
			const double k0 = calculate_square_2(p, this->p1, this->p2) / this->full_square;
			const double k1 = calculate_square_2(p, this->p2, this->p0) / this->full_square;
			const double k2 = calculate_square_2(p, this->p0, this->p1) / this->full_square;

			return transform([=, this]<class property_getter_t>(property_getter_t pg) -> auto { return pg(this->v0) * k0 + pg(this->v1) * k1 + pg(this->v2) * k2; });
		}

		[[nodiscard]] variable_t interpolate_point(RendererApi::PointF<2>::component_t x, RendererApi::PointF<2>::component_t y) const {
			return this->interpolate_point(RendererApi::PointF<2>{x, y});
		}
	};


	template<class variable_t>
	struct triangle_barycentric_interpolator {
	private:
		_triangle_barycentric_interpolator<variable_t, []<class interpolator_t>(interpolator_t interpolator) {
			return static_cast<variable_t>(interpolator([](variable_t v) { return v; }));
		}> impl;
	public:
		triangle_barycentric_interpolator(
				RendererApi::PointF<2> p0, variable_t v0,
				RendererApi::PointF<2> p1, variable_t v1,
				RendererApi::PointF<2> p2, variable_t v2
		) : impl{p0, v0, p1, v1, p2, v2} {};

		[[nodiscard]] variable_t interpolate_point(RendererApi::PointF<2> p) const {
			return this->impl.interpolate_point(p);
		}

		[[nodiscard]] variable_t interpolate_point(RendererApi::PointF<2>::component_t x, RendererApi::PointF<2>::component_t y) const {
			return this->impl.interpolate_point(x, y);
		}
	};

	template<>
	struct triangle_barycentric_interpolator<RendererApi::Color> {
	private:
		_triangle_barycentric_interpolator<RendererApi::Color, []<class interpolator_t>(interpolator_t interpolator) {
			return RendererApi::Color{
					static_cast<RendererApi::Color::component_fast_t>(interpolator([](RendererApi::Color const &c) { return c.red; })),
					static_cast<RendererApi::Color::component_fast_t>(interpolator([](RendererApi::Color const &c) { return c.green; })),
					static_cast<RendererApi::Color::component_fast_t>(interpolator([](RendererApi::Color const &c) { return c.blue; }))
			};
		}> impl;
	public:
		inline triangle_barycentric_interpolator(
				RendererApi::PointF<2> p0, RendererApi::Color v0,
				RendererApi::PointF<2> p1, RendererApi::Color v1,
				RendererApi::PointF<2> p2, RendererApi::Color v2
		) : impl{p0, v0, p1, v1, p2, v2} {
		}


		[[nodiscard]] inline RendererApi::Color interpolate_point(RendererApi::PointF<2> p) const {
			return this->impl.interpolate_point(p);
		}

		[[nodiscard]] inline RendererApi::Color interpolate_point(RendererApi::PointF<2>::component_t x, RendererApi::PointF<2>::component_t y) const {
			return this->impl.interpolate_point(x, y);
		}
	};

	template<>
	struct triangle_barycentric_interpolator<RendererApi::Color::Transparent> {
	private:
		union _interpolators_t {
			triangle_barycentric_interpolator<RendererApi::Color> ignore_alpha;
			_triangle_barycentric_interpolator<RendererApi::Color::Transparent, []<class interpolator_t>(interpolator_t interpolator) {
				return RendererApi::Color::Transparent{
						static_cast<RendererApi::Color::component_fast_t>(interpolator([](RendererApi::Color::Transparent const &c) { return c.red; })),
						static_cast<RendererApi::Color::component_fast_t>(interpolator([](RendererApi::Color::Transparent const &c) { return c.green; })),
						static_cast<RendererApi::Color::component_fast_t>(interpolator([](RendererApi::Color::Transparent const &c) { return c.blue; })),
						static_cast<RendererApi::Color::component_fast_t>(interpolator([](RendererApi::Color::Transparent const &c) { return c.alpha; }))
				};
			}> transparent;
		} interpolators;
		bool is_alpha_const;
		RendererApi::Color::component_fast_t const_alpha_value;

		inline triangle_barycentric_interpolator(
				bool is_alpha_const,
				RendererApi::PointF<2> p0, RendererApi::Color::Transparent v0,
				RendererApi::PointF<2> p1, RendererApi::Color::Transparent v1,
				RendererApi::PointF<2> p2, RendererApi::Color::Transparent v2
		) : interpolators{is_alpha_const ? _interpolators_t{.ignore_alpha{p0, v0.without_alpha(), p1, v1.without_alpha(), p2, v2.without_alpha()}} : _interpolators_t{.transparent{p0, v0, p1, v1, p2, v2}}},
			is_alpha_const{is_alpha_const}, const_alpha_value{v0.alpha} {}

	public:
		inline triangle_barycentric_interpolator(
				RendererApi::PointF<2> p0, RendererApi::Color::Transparent v0,
				RendererApi::PointF<2> p1, RendererApi::Color::Transparent v1,
				RendererApi::PointF<2> p2, RendererApi::Color::Transparent v2
		) : triangle_barycentric_interpolator{
				v0.alpha == v1.alpha && v0.alpha == v2.alpha,
				p0, v0, p1, v1, p2, v2
		} {}

		[[nodiscard]] inline RendererApi::Color::Transparent interpolate_point(RendererApi::PointF<2> p) const {
			if (this->is_alpha_const)
				return this->interpolators.ignore_alpha.interpolate_point(p).with_alpha(this->const_alpha_value);
			else
				return this->interpolators.transparent.interpolate_point(p);
		}

		[[nodiscard]] inline RendererApi::Color::Transparent interpolate_point(RendererApi::PointF<2>::component_t x, RendererApi::PointF<2>::component_t y) const {
			return this->interpolate_point(RendererApi::PointF<2>{x, y});
		}
	};

	template<unsigned DIMENSIONS, class component_t>
	struct triangle_barycentric_interpolator<RendererApi::Point<DIMENSIONS, component_t>> {
	private:
		using point_t = RendererApi::Point<DIMENSIONS, component_t>;

		_triangle_barycentric_interpolator<point_t, []<class interpolator_t>(interpolator_t interpolator) {
			if constexpr (DIMENSIONS == 2)
				return point_t{
						static_cast<point_t::component_t>(interpolator([](point_t const &p) { return p.x; })),
						static_cast<point_t::component_t>(interpolator([](point_t const &p) { return p.y; }))
				};
			else
				return point_t{
						static_cast<point_t::component_t>(interpolator([](point_t const &p) { return p.x; })),
						static_cast<point_t::component_t>(interpolator([](point_t const &p) { return p.y; })),
						static_cast<point_t::component_t>(interpolator([](point_t const &p) { return p.z; }))
				};
		}> impl;
	public:
		triangle_barycentric_interpolator(
				RendererApi::PointF<2> p0, point_t v0,
				RendererApi::PointF<2> p1, point_t v1,
				RendererApi::PointF<2> p2, point_t v2
		) : impl{p0, v0, p1, v1, p2, v2} {
		}


		[[nodiscard]] point_t interpolate_point(RendererApi::PointF<2> p) const {
			return this->impl.interpolate_point(p);
		}

		[[nodiscard]] point_t interpolate_point(RendererApi::PointF<2>::component_t x, RendererApi::PointF<2>::component_t y) const {
			return this->impl.interpolate_point(x, y);
		}
	};

	class triangle_barycentric_shader : public Shader {
	private:
		triangle_barycentric_interpolator<RendererApi::Color::Transparent> interpolator;
	public:
		inline triangle_barycentric_shader(
				RendererApi::PointF<2> p0, RendererApi::Color::Transparent c0,
				RendererApi::PointF<2> p1, RendererApi::Color::Transparent c1,
				RendererApi::PointF<2> p2, RendererApi::Color::Transparent c2
		) noexcept: interpolator{p0, c0, p1, c1, p2, c2} {}


		[[nodiscard]] inline RendererApi::Color::Transparent get_color(RendererApi::PointF<2>::component_t x, RendererApi::PointF<2>::component_t y) const final {
			return this->interpolator.interpolate_point(x, y);
		}
	};

	template<class variable_t, auto transform>
	struct _triangle_perspective_interpolator {
	private:
		long double full_square;
		RendererApi::PointF<3> p0;
		RendererApi::PointF<3> p1;
		RendererApi::PointF<3> p2;
		variable_t v0;
		variable_t v1;
		variable_t v2;

		static long double calculate_square_2(
				RendererApi::PointF<2> p0,
				RendererApi::PointF<3> p1,
				RendererApi::PointF<3> p2
		) noexcept {
			return (p1.x - p0.x) * (p2.y - p0.y) - (p2.x - p0.x) * (p1.y - p0.y);
		}

	public:
		_triangle_perspective_interpolator(
				RendererApi::PointF<3> p0, variable_t v0,
				RendererApi::PointF<3> p1, variable_t v1,
				RendererApi::PointF<3> p2, variable_t v2
		) : p0{p0}, p1{p1}, p2{p2}, v0{v0}, v1{v1}, v2{v2}, full_square{calculate_square_2({p0.x, p0.y}, p1, p2)} {
		}

		[[nodiscard]] variable_t interpolate_point(RendererApi::PointF<2> p) const {
			const double k0 = calculate_square_2(p, this->p1, this->p2) / (this->full_square * this->p0.z);
			const double k1 = calculate_square_2(p, this->p2, this->p0) / (this->full_square * this->p1.z);
			const double k2 = calculate_square_2(p, this->p0, this->p1) / (this->full_square * this->p2.z);
			const double divider = k0 + k1 + k2;

			return transform([=, this]<class property_getter_t>(property_getter_t pg) -> auto { return (pg(this->v0) * k0 + pg(this->v1) * k1 + pg(this->v2) * k2) / divider; });
		}

		[[nodiscard]] variable_t interpolate_point(RendererApi::PointF<2>::component_t x, RendererApi::PointF<2>::component_t y) const {
			return this->interpolate_point(RendererApi::PointF<2>{x, y});
		}
	};


	template<class variable_t>
	struct triangle_perspective_interpolator {
	private:
		_triangle_perspective_interpolator<variable_t, []<class interpolator_t>(interpolator_t interpolator) {
			return static_cast<variable_t>(interpolator([](variable_t v) { return v; }));
		}> impl;
	public:
		triangle_perspective_interpolator(
				RendererApi::PointF<3> p0, variable_t v0,
				RendererApi::PointF<3> p1, variable_t v1,
				RendererApi::PointF<3> p2, variable_t v2
		) : impl{p0, v0, p1, v1, p2, v2} {};

		[[nodiscard]] variable_t interpolate_point(RendererApi::PointF<2> p) const {
			return this->impl.interpolate_point(p);
		}

		[[nodiscard]] variable_t interpolate_point(RendererApi::PointF<2>::component_t x, RendererApi::PointF<2>::component_t y) const {
			return this->impl.interpolate_point(x, y);
		}
	};

	template<>
	struct triangle_perspective_interpolator<RendererApi::Color> {
	private:
		_triangle_perspective_interpolator<RendererApi::Color, []<class interpolator_t>(interpolator_t interpolator) {
			return RendererApi::Color{
					static_cast<RendererApi::Color::component_fast_t>(interpolator([](RendererApi::Color const &c) { return c.red; })),
					static_cast<RendererApi::Color::component_fast_t>(interpolator([](RendererApi::Color const &c) { return c.green; })),
					static_cast<RendererApi::Color::component_fast_t>(interpolator([](RendererApi::Color const &c) { return c.blue; }))
			};
		}> impl;
	public:
		inline triangle_perspective_interpolator(
				RendererApi::PointF<3> p0, RendererApi::Color v0,
				RendererApi::PointF<3> p1, RendererApi::Color v1,
				RendererApi::PointF<3> p2, RendererApi::Color v2
		) : impl{p0, v0, p1, v1, p2, v2} {
		}


		[[nodiscard]] inline RendererApi::Color interpolate_point(RendererApi::PointF<2> p) const {
			return this->impl.interpolate_point(p);
		}

		[[nodiscard]] inline RendererApi::Color interpolate_point(RendererApi::PointF<2>::component_t x, RendererApi::PointF<2>::component_t y) const {
			return this->impl.interpolate_point(x, y);
		}
	};

	template<>
	struct triangle_perspective_interpolator<RendererApi::Color::Transparent> {
	private:
		union _interpolators_t {
			triangle_perspective_interpolator<RendererApi::Color> ignore_alpha;
			_triangle_perspective_interpolator<RendererApi::Color::Transparent, []<class interpolator_t>(interpolator_t interpolator) {
				return RendererApi::Color::Transparent{
						static_cast<RendererApi::Color::component_fast_t>(interpolator([](RendererApi::Color::Transparent const &c) { return c.red; })),
						static_cast<RendererApi::Color::component_fast_t>(interpolator([](RendererApi::Color::Transparent const &c) { return c.green; })),
						static_cast<RendererApi::Color::component_fast_t>(interpolator([](RendererApi::Color::Transparent const &c) { return c.blue; })),
						static_cast<RendererApi::Color::component_fast_t>(interpolator([](RendererApi::Color::Transparent const &c) { return c.alpha; }))
				};
			}> transparent;
		} interpolators;
		bool is_alpha_const;
		RendererApi::Color::component_fast_t const_alpha_value;

		inline triangle_perspective_interpolator(
				bool is_alpha_const,
				RendererApi::PointF<3> p0, RendererApi::Color::Transparent v0,
				RendererApi::PointF<3> p1, RendererApi::Color::Transparent v1,
				RendererApi::PointF<3> p2, RendererApi::Color::Transparent v2
		) : interpolators{is_alpha_const ? _interpolators_t{.ignore_alpha{p0, v0.without_alpha(), p1, v1.without_alpha(), p2, v2.without_alpha()}} : _interpolators_t{.transparent{p0, v0, p1, v1, p2, v2}}},
			is_alpha_const{is_alpha_const}, const_alpha_value{v0.alpha} {}

	public:
		inline triangle_perspective_interpolator(
				RendererApi::PointF<3> p0, RendererApi::Color::Transparent v0,
				RendererApi::PointF<3> p1, RendererApi::Color::Transparent v1,
				RendererApi::PointF<3> p2, RendererApi::Color::Transparent v2
		) : triangle_perspective_interpolator{
				v0.alpha == v1.alpha && v0.alpha == v2.alpha,
				p0, v0, p1, v1, p2, v2
		} {}

		[[nodiscard]] inline RendererApi::Color::Transparent interpolate_point(RendererApi::PointF<2> p) const {
			if (this->is_alpha_const)
				return this->interpolators.ignore_alpha.interpolate_point(p).with_alpha(this->const_alpha_value);
			else
				return this->interpolators.transparent.interpolate_point(p);
		}

		[[nodiscard]] inline RendererApi::Color::Transparent interpolate_point(RendererApi::PointF<2>::component_t x, RendererApi::PointF<2>::component_t y) const {
			return this->interpolate_point(RendererApi::PointF<2>{x, y});
		}
	};

	template<unsigned DIMENSIONS, class component_t>
	struct triangle_perspective_interpolator<RendererApi::Point<DIMENSIONS, component_t>> {
	private:
		using point_t = RendererApi::Point<DIMENSIONS, component_t>;

		_triangle_perspective_interpolator<point_t, []<class interpolator_t>(interpolator_t interpolator) {
			if constexpr (DIMENSIONS == 2)
				return point_t{
						static_cast<point_t::component_t>(interpolator([](point_t const &p) { return p.x; })),
						static_cast<point_t::component_t>(interpolator([](point_t const &p) { return p.y; }))
				};
			else
				return point_t{
						static_cast<point_t::component_t>(interpolator([](point_t const &p) { return p.x; })),
						static_cast<point_t::component_t>(interpolator([](point_t const &p) { return p.y; })),
						static_cast<point_t::component_t>(interpolator([](point_t const &p) { return p.z; }))
				};
		}> impl;
	public:
		triangle_perspective_interpolator(
				RendererApi::PointF<3> p0, point_t v0,
				RendererApi::PointF<3> p1, point_t v1,
				RendererApi::PointF<3> p2, point_t v2
		) : impl{p0, v0, p1, v1, p2, v2} {
		}


		[[nodiscard]] point_t interpolate_point(RendererApi::PointF<2> p) const {
			return this->impl.interpolate_point(p);
		}

		[[nodiscard]] point_t interpolate_point(RendererApi::PointF<2>::component_t x, RendererApi::PointF<2>::component_t y) const {
			return this->impl.interpolate_point(x, y);
		}
	};

	class triangle_perspective_shader : public Shader {
	private:
		triangle_perspective_interpolator<RendererApi::Color::Transparent> interpolator;
	public:
		inline triangle_perspective_shader(
				RendererApi::PointF<3> p0, RendererApi::Color::Transparent c0,
				RendererApi::PointF<3> p1, RendererApi::Color::Transparent c1,
				RendererApi::PointF<3> p2, RendererApi::Color::Transparent c2
		) noexcept: interpolator{p0, c0, p1, c1, p2, c2} {}


		[[nodiscard]] inline RendererApi::Color::Transparent get_color(RendererApi::PointF<2>::component_t x, RendererApi::PointF<2>::component_t y) const final {
			return this->interpolator.interpolate_point(x, y);
		}
	};
}