#pragma once

#include <cstdlib>
#include <bgtu/computer_graphics_lab_work/renderer_api/voxel_painter.hpp>
#include "color.hpp"

namespace BGTU::ComputerGraphicsLabWork::Utilities {
	template<std::size_t bytes_per_pixel, void (pack_color)(unsigned char *buffer, std::size_t offset, RendererApi::Color color) noexcept>
	class VoxelPainterExporterToPackedArray : public RendererApi::VirtualVoxelPainter::Exporter {
	public:
		RendererApi::VirtualVoxelPainter::visible_pixel_coordinate_fast_t w;
		RendererApi::VirtualVoxelPainter::visible_pixel_coordinate_fast_t h;
		unsigned char *pixels_packed;
		RendererApi::SpriteMetadata **pixel_owners;

		constexpr VoxelPainterExporterToPackedArray(
				RendererApi::VirtualVoxelPainter::visible_pixel_coordinate_fast_t w,
				RendererApi::VirtualVoxelPainter::visible_pixel_coordinate_fast_t h,
				unsigned char *pixels_packed,
				RendererApi::SpriteMetadata **pixel_owners
		) noexcept: w{w}, h{h}, pixels_packed{pixels_packed}, pixel_owners{pixel_owners} {}

		constexpr void set_pixel(RendererApi::PointI<2>::component_t x, RendererApi::PointI<2>::component_t y, RendererApi::Color c, RendererApi::SpriteMetadata *owner) noexcept final {
			assert(x < this->w);
			assert(y < this->h);
			std::size_t offset = y * this->w + x;
			this->pixel_owners[offset] = owner;
			pack_color(this->pixels_packed, offset * bytes_per_pixel, c);
		};
	};

	using VoxelPainterExporterToBgr32Array = VoxelPainterExporterToPackedArray<4, PackColor::color_to_bgr32>;
}