#pragma once

#include <cstdlib>
#include <exception>
#include <bgtu/computer_graphics_lab_work/renderer_api/point.hpp>
#include <bgtu/computer_graphics_lab_work/renderer_api/voxel_painter.hpp>
#include <bgtu/computer_graphics_lab_work/renderer_api/sprite.hpp>
#include "shapes/rectangle.hpp"

namespace BGTU::ComputerGraphicsLabWork::Utilities {
	template<class unzoomed_painter_t =  RendererApi::VirtualVoxelPainter>
	class ZoomedVoxelPainter : public RendererApi::VirtualVoxelPainter {
	private:
		unzoomed_painter_t *dst;
		RendererApi::PointI<2>::component_t cx, cy;
		std::size_t pixel_size;
	public:
		constexpr ZoomedVoxelPainter(
				unzoomed_painter_t *dst,
				RendererApi::PointI<2>::component_t cx,
				RendererApi::PointI<2>::component_t cy,
				std::size_t pixel_size
		) noexcept: dst{dst}, cx{cx}, cy{cy}, pixel_size{pixel_size} {}

		constexpr ZoomedVoxelPainter(
				unzoomed_painter_t *dst,
				RendererApi::PointI<2> center_tl,
				std::size_t pixel_size
		) noexcept: dst{dst}, cx{center_tl.x}, cy{center_tl.y}, pixel_size{pixel_size} {}

		RendererApi::SpriteMetadata *current_owner = nullptr;

		void add_voxel(long long x, long long y, double z, RendererApi::Color::Transparent c) final {
			Shapes::fill_square<unzoomed_painter_t>(
					this->dst,
					this->cx + x * this->pixel_size,
					this->cy + y * this->pixel_size,
					this->pixel_size,
					z,
					c,
					current_owner
			);
		}

		void add_voxel(long long x, long long y, double z, RendererApi::Color::Transparent c, RendererApi::SpriteMetadata *owner) final {
			Shapes::fill_square<unzoomed_painter_t>(
					this->dst,
					this->cx + x * this->pixel_size,
					this->cy + y * this->pixel_size,
					this->pixel_size,
					z,
					c,
					owner
			);
		}

		[[nodiscard]] constexpr visible_pixel_coordinate_fast_t width() const override {
			throw std::exception{"This wrapper doesn't provides such information"};
			return 0;
		}

		[[nodiscard]] constexpr visible_pixel_coordinate_fast_t height() const override {
			throw std::exception{"This wrapper doesn't provides such information"};
			return 0;
		}
	};
}